/*
  Represent an audio file with a set of voices speaking names
  Copyright (C) 2013 Neil Birkbeck

  This library is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation; either version 2.1 of the License, or (at your option)
  any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License along
  with this library; if not, write to the Free Software Foundation, Inc., 59
  Temple Place, Suite 330, Boston, MA 02111-1307 USA

*/
#include <stdio.h>
#include <stdlib.h>

#include <vector>
#include <iostream>
#include <fstream>

#include <GL/gl.h>
#include <nmisc/commandline.h>
#include <nmath/matrix.h>


// Convenience class for loading single channel PCM audio into a buffer.
// Pretty much just a vector; Doesn't store any information about the 
// sampling rate, etc.
class AudioSample {
public:
  AudioSample() { }

  // Initialize a buffer of the given length with value v.
  AudioSample(size_t length, float v): data(length, v) { }

  // Load filename, return true on success.
  bool Load(const std::string& filename) {
    data.clear();

    FILE* file = fopen(filename.c_str(), "rb");
    if (!file) {
      fprintf(stderr, "ERROR: unable to open file %s for read\n", 
	      filename.c_str());
      return false;
    }

    float buffer[1024];
    size_t nread = 0;

    while ((nread = fread(buffer, sizeof(float),  1024, file)) > 0) {
      data.insert(data.end(), buffer, buffer + nread);
      if (nread < 1024) break;
    }

    printf("INFO: read %s %d\n", filename.c_str(), data.size());

    fclose(file);    
    return true;
  }

  // Write out the buffer to the filename; return true on success.
  bool Write(const std::string& filename) {
    FILE* file = fopen(filename.c_str(), "wb");
    if (!file) return false;

    fwrite(&(data[0]), sizeof(float), data.size(), file);

    fclose(file);
    return true;
  }

  const float& operator[](const int i) const { return data[i];}

  float& operator[](const int i) { return data[i]; }
  
  size_t size() const { return data.size(); }

protected:
  std::vector<float> data;
};


// Structure representing the positional and scale information regarding
// a fit of an audio sample from the database to a probe audio sample.
struct Fit {  
  bool operator<(const Fit& e) const {
    return error < e.error;
  }
  
  bool operator<=(const Fit& e) const {
    return error <= e.error;
  }

  double error; //!< Error of the fit.
  int index;    //!< Index in the database.
  int location; //!< The positional location of the fit.
  double scale; //!< How much to scale the .
};


// Simple structure representing a segment (or range) of a file.
struct Segment {
  int start, end;
}; 


// A database of audio samples, with the ability to fit the audio samples
// to a probe.
class AudioDatabase {
public:

  // Load the AudioDatabase from a directory.  The directory should contain
  // a list of the relative paths to the audio samples.
  bool Load(const std::string& dir) {
    const std::string listing = dir + "/list.txt";
    std::ifstream ifstr(listing.c_str());

    if (!ifstr) return false;
    
    std::string filename;
    while (ifstr) {
      ifstr >> filename;
      const std::string f = dir + "/" + filename;

      AudioSample sample;
      if (!sample.Load(f)) {
	continue;
      }
      samples.push_back(sample);
    }    
    return samples.size() > 0;
  }

  // Approximate the probe sample with samples from the database using maxits
  // iterations.  On output the probe will be replaced with its approximation.
  //
  // On each iteration, every sample in the database is least-squares fit to
  // almost every position in the probe file.  Then the "fits" with the least
  // error are chosen (without overlap) and added to an accumulator.  In this
  // way, each iteration corresponds to approximating the sound with a layer 
  // of non-overlapping audio samples.
  bool Approximate(AudioSample& probe, const int maxits) {
    AudioSample approx(probe.size(), 0.0);

    for (int i = 0; i < maxits; ++i) {
      std::vector<Segment> segments;
      std::vector<Fit> all_fits;

      if (!FindAllFitErrors(probe, approx, all_fits)) {
	fprintf(stderr, 
		"ERROR: unable to find best location on iteration: %d\n", i);
	return false;
      }

      // Sort the fits by increasing error.
      std::sort(all_fits.begin(), all_fits.end());
      
      // Start approximating the signal with the best errors.
      // Only keep those fits that do not overlap with previous selected
      // ones.  Using non-overlapping segments assures that fit.error is
      // correct and does not need to be recomputed.
      for (int j = 0; j < all_fits.size(); ++j) {
	Segment segment;
	segment.start = all_fits[j].location;
	segment.end = all_fits[j].location + 
	  samples[all_fits[j].index].size();

	if (Intersects(segments, segment)) {
	  continue;
	}

	AddToApprox(approx, all_fits[j]);
	AddSegment(segments, segment);
      }
    }

    probe = approx;
    return true;
  }

  // Find fit errors for each sample in the database and return the results
  // in the all_fits vector.
  bool FindAllFitErrors(const AudioSample& probe,
			const AudioSample& approx,
			std::vector<Fit>& all_fits) {
    all_fits.clear();

    // Don't actually sample every possible offset, but only every nth.
    const int sample_rate = 500;
    double best_error = 1e10;
    int best_index = 0;
    int best_location = 0;
    
    for (int t = rand() % sample_rate; t < probe.size(); t += sample_rate) {
      if (t % 1000 == 0) printf("%d\n", t);

      for (int k = 0; k < samples.size(); ++k) {
	if (t + samples[k].size() <= probe.size()) {
	  double scale = 0;
	  const double error = FindBestFit(t, probe, approx, k, scale);

	  Fit fit;
	  fit.error = error;
	  fit.index = k;
	  fit.location = t;
	  fit.scale = scale;

	  if (scale > 0) {
	    all_fits.push_back(fit);
	  }
	}
      }
    }
    return all_fits.size() != 0;
  }

  // Find the best scale of a database sample, k, to fit the residual between
  // the probe and current approx audio sample at time t in the probe.
  // Puts the resulting scale in the output parameter scale and returns the 
  // residual.
  double FindBestFit(const int t,
		     const AudioSample& probe,
		     const AudioSample& approx,
		     const int k,
		     double& scale) {
    // The system of equations is:
    //  db_i * scale = (probe_{t + i} - approx_{t + i}
    // Since db_i is just a vector we can solve the normal equations
    // with dot products.
    double AtA = 0;
    double Atb = 0;
    for (int i = 0; i < samples[k].size(); ++i) {
      double d = probe[t + i] - approx[t + i];
      AtA += samples[k][i] * samples[k][i];
      Atb += samples[k][i] * d;
    }
    scale = Atb/AtA;

    double error = 0;
    for (int i = 0; i < samples[k].size(); ++i) {
      double d = (probe[t + i] - (approx[t + i] + samples[k][i]*scale));
      error += d*d;
    }
    return error;
  }

  // Add the "fit" to the current approximation signal.
  void AddToApprox(AudioSample& approx, const Fit& fit) {
    for (int i = 0; i < samples[fit.index].size(); ++i) {
      approx[i + fit.location] += samples[fit.index][i]*fit.scale;
    }
  }

  // Return true if the "segment" intersects any of the "segments".
  bool Intersects(const std::vector<Segment>& segments, 
		  const Segment& segment) {
    for (int i = 0; i < segments.size(); ++i) {
      if ((segments[i].start < segment.start && 
	   segments[i].end >= segment.start) ||
	  (segments[i].start < segment.end && 
	   segments[i].end >= segment.end)) {
	return true;
      }
    }
    return false;
  }
  
  // Add a segment to the list of segments.  Merge the segments if possible.
  void AddSegment(std::vector<Segment>& segments, 
		  const Segment& segment) {
    int insertLocation = segments.size();
    for (int i = 0; i < segments.size(); ++i) {
      if (segments[i].start >= segment.start) {
	insertLocation = i; // TODO(this was a bug at zero before).
	break;
      }
    }
    if (insertLocation == segments.size()) {
      segments.push_back(segment);
    } else {
      // We can merge the two if they overlap.  The closed segments shouldn't
      // really overlap but rather meet at the boundaries.
      if (segment.end >= segments[insertLocation].start) {
	segments[insertLocation].start = segment.start;
	if (segment.end > segments[insertLocation].end) 
	  segments[insertLocation].end = segment.end;
      } else {
	segments.insert(segments.begin() + insertLocation, segment);
      }
      
      // See if we overlap with the previous segment.  If so we can merge again.
      if (insertLocation > 0 && 
	  segments[insertLocation - 1].end >= segments[insertLocation].start) {
	segments[insertLocation - 1].end = 
	  std::max(segments[insertLocation - 1].end, segments[insertLocation].end);

	segments.erase(segments.begin() + insertLocation);
      }
    }
  }

protected:
  std::vector<AudioSample>  samples;
};


int main(int ac, char* av[]) {
  nacb::CommandLine cline;
  std::string dbdir, probename, output =  "/tmp/out.raw";
  int its = 10;

  cline.registerOption("db", "Location of the database folder.", &dbdir);
  cline.registerOption("probe", "The probe folder.", &probename);
  cline.registerOption("out", "The output file", &output);
  cline.registerOption("its", "Iterations", &its);
  cline.parse(ac, av);

  AudioDatabase db;
  if (!db.Load(dbdir)) {
    std::cerr << "Unable to load database from:" << dbdir << "\n";
    return -1;
  }

  AudioSample sample;
  if (!sample.Load(probename)) {
    std::cerr << "Unable to load probe from:" << probename << "\n";
    return -2;
  }

  // Approximate the audio sample with the database; use a certain
  // number of iterations.  
  db.Approximate(sample, its);

  sample.Write(output.c_str());
  
  return 0;
}
